/**********************************************************************/
// 3x3matrix.v - 3X3 Matrix Multiply 
//               Implementation using basic equations 
// 
//                  Author: Latha  Pillai 
//                  Senior Applications Engineer 
//                  Video Applications 
//                  Advanced Products Division 
//                  Xilinx, Inc. 
// 
//                  Copyright (c) 1999 Xilinx, Inc. 
//                  All rights reserved 
// 
//                  Date:   Aug. 7, 2001 
//                  For:    Application note XAPP284 
// 
//                  RESTRICTED RIGHTS LEGEND 
// 
//      This software has not been published by the author, and 
//      has been disclosed to others for the purpose of enhancing 
//      and promoting design productivity in Xilinx products. 
// 
//      Therefore use, duplication or disclosure, now and in the 
//      future should give consideration to the productivity 
//      enhancements afforded the user of this code by the author's 
//      efforts.  Thank you for using our products ! 
// 
// Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
//              WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
//              IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR 
//              A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT. 
// 
// Revision: 
//          Aug. 7, 2001     Creation 
//          Aug. 7, 2001     Rev1. Gregg  Hawkes 
// 
// 
// Other modules instanced in this design: 
// 
//          MULT18X18 
/* 

BRIEF DESCRIPTION 

This code describes using a technique called time multiplexing to 
leverage a fast hardware multiply in a relatively slow operation, 
thereby increasing the efficiency of the implementation. 

The operation being shown is a 3X3 matrix of constants times a 3 
component vector. The equations look like: 

KA1 * A + KA2 * A + KA3 * A = X 
KB1 * A + KB2 * A + KB3 * A = X 
KC1 * A + KC2 * A + KC3 * A = X 

The hardware to accomplish this task consists of a multiplier fed by 3 
input registers and an accumulator to compute the three terms in each 
line above. 
  

DETAILED DESCRIPTION: 

The multiplier output is fed into the adder A input. It takes 3 clk 
cycles for the first valid mutiplier output reach the adder input A. The 
B input of the adder can be a zero or the adder's accumulating register. 
By selecting a zero on the B input the adder just passes the input A 
through to the accumulating register. By selecting the accumulating 
register, the contents of the previous add can be added to the output of 
the multiplier. 

The repeating flow for the accumlating register will be for the 1st clk, 
the mux output is '0', so we always pass the first argument through to 
the accumulator register. For the 2nd and 3rd clks, the accumulator 
register is fed back and added to the output of the multiply. This is 
made possible by using the cntr3 outputs as the select lines. 

The following text describes the condition of the internal nodes after 
consecutive clocks.  The state of the nodes assumes the clock has 
occured and data is stable. 
  

clock    multiplier     adder               adder 
number   output         output              output 
         register                           register 
--------------------------------------------- 
rst      X              X                   0 
1        X              X                   0 

2        X              X                   0 
3        KA1*A          KA1*A               0 
4        KB1*B          KA1*A+KB1*B         KA1*A 
5        KC1*C          KA1*A+KB1*B+KC1*C   KA1*A+KB1*B 
6        KA2*A          KA2*A               KA1*A+KB1*B+KC1    answer 1 
7        KB2*B          KA2*A+KB2*B         KA2*A 
8        KC2*C          KA2*A+KB2*B+KC2*C   KA2*A+KB2*B 
9        KA3*A          KA3*A               KA2*A+KB2*B+KC2*C  answer 2 
10       KB3*B          KA3*A+KB3*B         KA3*A 
11       KC3*C          KA3*A+KB3*B+KC3*C   KA3*A+KB3*B 
12       next KA1*A     next KA1*A          KA3*A+KB3*B+KC3*C  answer 3 
*/
/***********************************************************************/

module matrix3x3 ( CLK, RST, CWEL, A, B, C, KA, KB, KC, X, Y, Z );

output [11:0] X, Y, Z;
input CLK, RST;
input[11:0] A, B, C;
input[9:0] KA, KB, KC;
input[1:0] CWEL ;

reg[11:0] A_reg, B_reg, C_reg, A_reg1, B_reg1, C_reg1;
//reg[11:0] A_reg_int, B_reg_int, C_reg_int;
reg[11:0] X, Y, Z;
reg[11:0] data_mux;
reg [9:0] KA1, KB1, KC1, KA2, KB2, KC2, KA3, KB3, KC3;
reg[9:0] coeff_mux;
reg[1:0] cntr3,cntr3_en;
reg[3:0] cntr9,cntr9_out;
reg[35:0] P1_reg;
reg[35:0] adder_mux,sum;
wire[35:0] P1;//, P2, P3;
reg wait1;
reg[1:0] i_wait;
reg[2:0] cnt9_wait;
reg[1:0] CWEL_reg, cntr_cwel;
integer indexi,i,j;

/* ----------DATA INPUT SECTION------------- */

/* In the 3:1 data mux. To match the pipeline of the
Data inputs with the coeeficient inputs, the data values are registered first
At the input of the 3:1 mux and then again at the output of the 3:1 mux.
To make sure that the inputs don't change in the middle of a set of vector
summation, the inputs are held constant for 9 clks. This will ensure that 
the input values seen by the 3x3 vector is the same for the first set of
answers. */

/* cntr3 to count 0-1-2-3-1-2-3 */
always @ (posedge CLK or posedge RST) 
begin 
  if (RST) cntr3 <= 2'b00; 
  else //if (cnt9_wait == 1'b0)
     begin
     if (cntr3 == 2'b11) cntr3 <= 2'b01; 
        else cntr3 <= cntr3 + 1; 
     end 
end

/* inputs registered twice to match the pipe length of the coeeficients */

always @ (posedge CLK or posedge RST)
   begin
   if (RST)
      begin
	A_reg1 <= 12'b0; B_reg1 <= 12'b0; C_reg1 <= 12'b0;
      end
   else if (j == 0)
      begin 
         A_reg1 <= A; B_reg1 <= B; C_reg1 <= C;
      end
   end

always @ (posedge CLK or posedge RST)
   begin
   if (RST)
      begin
	A_reg <= 12'b0; B_reg <= 12'b0; C_reg <= 12'b0;
      end
   else
      begin 
         A_reg <= A_reg1; B_reg <= B_reg1; C_reg <= C_reg1;
      end
   end

always @ (posedge CLK or posedge RST)
   begin
   if (RST)
      begin 
        j = 0; 
      end
   else 
      begin
      if ( j < 8)
         begin
           j = j + 1;
         end
      else j <= 0;
      end
   end


/* ----------MODE SELECT SECTION------------- */

/* mode select should be constant for 3 clk cycles  to complete one set
of coefficients. So modeselect is updated every 3rd clk */


/* cntr3 used to hold CWEL constant for 3 clocks. */
always @ (posedge CLK or posedge RST)
   begin
   if (RST)
      begin
       CWEL_reg <= 2'b00; i <= 0;
      end
   else  
       begin           
          CWEL_reg <= CWEL;
          for (i = 0;i < 3 ;i = i+1)
              begin
              end
       end
   end

/* coefficient register update. The register shd hold the
value for 3 clks to get the right output. */
always @ (posedge CLK or posedge RST)
   begin
   if (RST)
      begin
       KA1 <= 0; KB1 <= 0; KC1 <= 0;
       KA2 <= 0; KB2 <= 0; KC2 <= 0;
       KA3 <= 0; KB3 <= 0; KC3 <= 0;
      end
   else
       begin
           case (CWEL_reg)
           2'b01: begin KA1 <= KA; KB1 <= KB; KC1 <= KC; end
           2'b10: begin KA2 <= KA; KB2 <= KB; KC2 <= KC; end
	         2'b11: begin KA3 <= KA; KB3 <= KB; KC3 <= KC; end
           2'b00: begin KA1 <= KA1; KB1 <= KB1; KC1 <= KC1;
                        KA2 <= KA2; KB2 <= KB2; KC2 <= KC2;
                        KA3 <= KA3; KB3 <= KB3; KC3 <= KC3; end
	         endcase
       end
   end

/* ----------COEFFIECIENT MUX SECTION------------- */

/*cntr9 to count 0-1-2-3-4-5-6-7-8-9-1 */

always @ (posedge CLK or posedge RST) 
begin 
  if (RST) cntr9 <= 4'b0; 
  else 
  begin
     if (cntr9 == 4'b1001) 
         cntr9 <= 4'b0001; 
     else cntr9 <= cntr9 + 1; 
  end
end 

always @ (posedge RST or posedge CLK)
   begin
   if (RST)
       begin
       coeff_mux <= 0; data_mux <= 0; 
       end
   else
       begin
	     case (indexi)
         0 : begin coeff_mux <= KA1; data_mux <= A_reg;end
         1 : begin coeff_mux <= KB1; data_mux <= B_reg;end
         2 : begin coeff_mux <= KC1; data_mux <= C_reg;end      
         3 : begin coeff_mux <= KA2; data_mux <= A_reg;end
         4 : begin coeff_mux <= KB2; data_mux <= B_reg;end
         5 : begin coeff_mux <= KC2; data_mux <= C_reg;end
         6 : begin coeff_mux <= KA3; data_mux <= A_reg;end
         7 : begin coeff_mux <= KB3; data_mux <= B_reg;end
         8 : begin coeff_mux <= KC3; data_mux <= C_reg;end
      endcase
      end
   end

always @ (posedge CLK or posedge RST) 
begin 
  if (RST) 
    begin i_wait <= 2'b01; end
  else if (i_wait > 2'b00)
    begin i_wait <= i_wait - 1; end
  else i_wait <= i_wait;
end 

always @ (posedge RST or posedge CLK)
  begin
    if (RST)
      begin
        indexi <= 8;
      end
    else if (i_wait == 2'b00)
      begin
       if (indexi == 8)
          indexi <= 0;
       else
          indexi <= indexi + 1;
      end
  end

/* ----------MULTIPLIER SECTION------------- */

/* 9x pumped multiplier; P3 registered twice to match the pipelining
   of the first adder  */
   
MULT18X18 U1_MULT18X18 ( .A({8'b0,coeff_mux}), .B({6'b0,data_mux}), .P(P1)); 

/* registering multiplier outputs */
always @ (posedge CLK or posedge RST)
   begin
   if (RST)
      begin P1_reg <= 0; end
   else 
      begin P1_reg <= P1; end
   end


/* ----------ADDER SECTION------------- */

/* Adder mux. Inputs a '0' every 3rd clk */
always @ (cntr3[1] or cntr3[0] or sum)
begin
adder_mux = ((cntr3[1]==1'b 0) & (cntr3[0] ==1'b 1)) ? 0 : sum;
end

/* Final adder */
always @ (posedge CLK or posedge RST)
   if (RST)
      begin
       sum <= 0; 
      end
   else  
      begin
      sum <= P1_reg + adder_mux ;
   end


/* ----------OUTPUT SECTION------------- */


/* At the output of the adder, the first valid X values appears at the 6th clk
after reset. After this, at every 3rd clk, a valid output values are obtained for
Y ,Z, X, Y, Z and so on.  This function is realised using a enable cntr. The cntr 
after reset, counts upto 3 at which point another output counter is enabled. The
output of the enable counter holds its value of 3 as long as it is not reset. */

/* output cntr starts after 4 clk to match the initial pipe
delays of inputs/coeeficients */

always @ (posedge CLK or posedge RST) 
begin 
  if (RST) 
    begin cnt9_wait <= 3'b100; end
  else if (cnt9_wait > 3'b000)
    begin cnt9_wait <= cnt9_wait - 1; end
  else cnt9_wait <= cnt9_wait;
end 
/*cntr9_out to count 0-1-2-3-4-5-6-7-8-9-1-2- */

 always @ (posedge CLK or posedge RST) 
    begin 
       if (RST) cntr9_out <= 4'b0000; 
       else if (cnt9_wait == 3'b000)
         begin 
           if(cntr9_out == 4'b1001) cntr9_out <= 4'b0001; 
           else cntr9_out <= cntr9_out + 1; 
        end 
    end

/* adder output assigned to X,Y and Z */
always @ (posedge CLK or posedge RST)
   begin
   if (RST)
      begin
      X <= 0; Y <= 0; Z <= 0;
      end
   else 
      begin
      case (cntr9_out)
           4'b0001: begin X <= X; Y <= Y; Z <= Z; end
           4'b0010: begin X <= X; Y <= Y; Z <= Z; end
           4'b0011: begin X <= sum[11:0]; Y <= Y; Z <= Z; end
           
           4'b0100: begin X <= X; Y <= Y; Z <= Z; end
           4'b0101: begin X <= X; Y <= Y; Z <= Z; end
           4'b0110: begin X <= X; Y <= sum[11:0]; Z <= Z; end
	     
           4'b0111: begin X <= X; Y <= Y; Z <= Z; end
           4'b1000: begin X <= X; Y <= Y; Z <= Z; end
           4'b1001: begin X <= X; Y <= Y; Z <= sum[11:0]; end
           default: ;
	   endcase
       end
   end
endmodule

module MULT18X18 (A, B, P); // synthesis syn_black_box
input [17:0]A;
input [17:0]B;
output [35:0]P;


endmodule

